
#
# COMP 304 2009
#
# 2007 Alexandre Denault
# 2008, 2009 small updates by Hans Vangheluwe
#

from Tkinter import *

#####
# The GridWindow class can be used to generate Sudoku grids of
# various sizes.
class GridWindow:
    
    # Creates a sudoko grid of the specified size
    def __init__(self, master, rows, cols, 
                               rowOffset=0, colOffset=0,
                               rowHeight=40, colWidth=40,
                               squareColourInFocus="LightGray",
                               squareColourOutFocus="White",
                               dividerWidth=3):
        
        self.rows = rows
        self.cols = cols
       
        self.rowHeight = rowHeight
        self.colWidth  = colWidth

        self.squareColourInFocus = squareColourInFocus
        self.squareColourOutFocus = squareColourOutFocus

        self.dividerWidth = dividerWidth
 
        self.master = master
        self.window = Toplevel(master) # create a separate window
        self.window.title("("+str(rowOffset)+","+str(colOffset)+")")

        # bind UI events to callback methods
        self.window.bind("<Button-1>", self.onClick)
        self.window.bind("<Key>", self.onKey)
        self.window.protocol("WM_DELETE_WINDOW", self.onClose)

        # the grid is drawn on a Tkinter Canvas
        self.canvas = Canvas(self.window, width=cols*self.colWidth+2, height=rows*self.rowHeight+2)
        self.canvas.pack()
               
        # containers of the GridWindow elements 
        self.rectangle = {}
        self.text = {}
        
        # The coordinate of the square that currently has focus. 
        # A square's coordinate is stored as a tuple (row, col),
        # where row and col are integers in the range 0 to 9
        self.focus = None 
        
        # These are the keyboard keys that can be used to clear a
        # square. Right now, these are space, backspace and enter.
        # self.clearkey = [13, 8, 32]
        # Updated March 2009 based on
        #  http://infohost.nmt.edu/tcc/help/pubs/tkinter/key-names.html 
        #  and a simple experiment printing out event.keycode
        #  keycode(SPACE)     = 65
        #  keycode(BACKSPACE) = 22
        #  keycode(ENTER)     = 36
        self.clearkey = [65, 22, 36]
    
        for row in range(0, self.rows):
            for col in range(0, self.cols):
                self.rectangle[(row, col)] = \
                  self.canvas.create_rectangle(2+colWidth*col, 2+rowHeight*row, 2+colWidth*(col+1), 2+rowHeight*(row+1), fill="White")   
                self.text[(row, col)] = \
                  self.canvas.create_text(2+colWidth*(col+0.5), 2+rowHeight*(row+0.5), text=str((row,col)))
                  # show coordinates in cells
                  # note how on a Tkinter Canvas, (x,y) coordinate (0,0) is the top left corner
    
    # Sets a number in the specified square  
    def setNumber(self, location, number):
        itemId = self.text[location]
        self.canvas.itemconfigure(itemId, text=number)
    
    # Sets the colour of the specified square
    def setColour(self, location, colour):        
        itemId = self.rectangle[location]
        self.canvas.itemconfigure(itemId, fill=colour)
    
    # Called when a square is clicked on
    # Sets its background (fill) colour
    def onClick(self, event):        
        row = (event.y - 2) / self.rowHeight 
        col = (event.x - 2) / self.colWidth
                
        if row >= 0 and row < self.rows and col >= 0 and col < self.cols:
            if (self.focus != None):
                self.setColour(self.focus, self.squareColourOutFocus)                
            self.focus = (row, col)
            self.setColour(self.focus, self.squareColourInFocus)
    
    # Called when a keyboard key is pressed 
    # (when this Sudoku window has the focus).
    def onKey(self, event):
        if self.focus != None:
            if "123456789".find(event.char) != -1:
                newValue = event.char            
            elif event.keycode in self.clearkey:
                newValue = ""                
            else:
                return
            
            # Here, we are directly setting updating what is displayed in the grid.
            self.setNumber(self.focus, newValue)

            # For assignment 5 you should use the Observer Pattern.
            # This means the state should be kept centrally and
            # any change to the central state should be reflected in all the appropriate views.
    
    # Called then a particular window is closed.
    def onClose(self):
        self.window.destroy()
            
class MainWindow(GridWindow):
    
    def __init__(self, master):
        
        GridWindow.__init__(self, master,9,9)
       
        for i in range(0,2):
            # thick vertical dividing line every three squares
            self.canvas.create_line(2+3*self.colWidth*(i+1), 2, 
                                    2+3*self.colWidth*(i+1), 2+9*self.rowHeight, 
                                    width=self.dividerWidth)  
            # thick horizontal dividing line every three squares
            self.canvas.create_line(2, 2+3*self.rowHeight*(i+1), 
                                    2+9*self.colWidth, 2+3*self.rowHeight*(i+1), 
                                    width=self.dividerWidth)

#####
# Creates a 1 by 9 Sudoku grid, suitable for displaying a row.
class RowWindow(GridWindow):
    
    def __init__(self, master, rowOffset, colOffset):
        
        GridWindow.__init__(self, master, 1, 9, rowOffset, colOffset)

#####
# Creates a 9 by 1 Sudoku grid, suitable for displaying a column.        
class ColWindow(GridWindow):
    
    def __init__(self, master, rowOffset, colOffset):
        
        GridWindow.__init__(self, master, 9, 1, rowOffset, colOffset)

#####
# Creates a 3 by 3 Sudoku grid, suitable for displaying a block.
class BlockWindow(GridWindow):
    
    def __init__(self, master, rowOffset, colOffset):
        
        GridWindow.__init__(self, master, 3, 3, rowOffset, colOffset)
                
